import logging

from komora_syncer.config import get_config
from komora_syncer.connections.NetboxConnection import NetboxConnection
from komora_syncer.models.netbox.NetboxBase import NetboxBase

logger = logging.getLogger(__name__)


class NbInventoryItem(NetboxBase):
    def __init__(self, komora_obj):
        NetboxBase.__init__(self)
        self.inventory_number = komora_obj.inventarizationNumber
        self.komora_id = komora_obj.id
        self.komora_site_id = komora_obj.siteId
        self.komora_location_id = komora_obj.locationId
        self.komora_device_id = komora_obj.deviceId
        self.komora_url = f"{get_config()['komora']['KOMORA_URL']}/app/inventarization/{self.inventory_number}/{self.komora_device_id}"
        self.netbox_url = komora_obj.netboxUrl
        self.api_object = None

    def find(self):
        if not self.komora_id:
            return self.api_object

        netbox_inventory_item = self._find_by_komora_id()
        if netbox_inventory_item:
            self.api_object = netbox_inventory_item
            return self.api_object

        netbox_inventory_item = self._find_by_inventory_number()
        if netbox_inventory_item:
            self.api_object = netbox_inventory_item
            return self.api_object

        logger.warning("Unable to find Inventory Item")
        return self.api_object

    def _find_by_komora_id(self):
        return NetboxConnection.get_connection().dcim.inventory_items.get(
            cf_komora_id=self.komora_id)

    def _find_by_inventory_number(self):
        return NetboxConnection.get_connection().dcim.inventory_items.get(
            cf_asset_numbers=self.inventory_number)

    def update(self, nb_inventory_item):
        try:
            if nb_inventory_item.update(self.params):
                self.api_object = nb_inventory_item
                logger.info(
                    f"Inventory: {self.inventory_number} updated successfuly")
        except Exception as e:
            logger.exception(
                f"Unable to update Inventory Item: {self.inventory_number}")

    def synchronize(self):
        inventory_item = self.find()

        if inventory_item:
            self.update(inventory_item)
        else:
            logger.info(
                f"Inventory Item: {self.inventory_number} not found in Netbox.")

    @property
    def params(self):
        params = {}

        if self.api_object:
            #TODO: Workaround for multiple asset_numbers per item
            if ";" in self.api_object.custom_fields["asset_numbers"] and self.inventory_number in self.api_object.custom_fields["asset_numbers"].split(";"):
                params['custom_fields'] = self.api_object.custom_fields
                params['custom_fields']['komora_id'] = self.api_object.custom_fields["komora_id"]
                params['custom_fields']['komora_url'] = self.api_object.custom_fields["komora_url"]
                logger.debug(params)

            elif isinstance(self.api_object.custom_fields, dict):
                params['custom_fields'] = self.api_object.custom_fields
                params['custom_fields']['komora_id'] = self.komora_id
                params['custom_fields']['komora_url'] = self.komora_url
        else:
            params['custom_fields'] = {
                "komora_id": self.komora_id, "komora_url": self.komora_url}

        return params

    def get_nb_inventory_items_data():
        # TODO: WORKAROUND
        SITES_CACHE = {
            site.id: site for site in NetboxConnection.get_connection().dcim.sites.all()}
        DEVICES_CACHE = {dev.id: {
            "device": dev, "site": SITES_CACHE[dev.site.id], "location": dev.location} for dev in NetboxConnection.get_connection().dcim.devices.all()}

        def inventory_to_json(inventory):
            device_id = inventory.device.id
            device = DEVICES_CACHE[device_id]['device']
            site = DEVICES_CACHE[device_id]['site']
            location = DEVICES_CACHE[device_id]['location']

            # print(inventory)
            inventory_dict = {}
            inventory_dict['serial'] = inventory.serial
            # inventory_dict['last_probed'] = inventory.custom_fields.get(
            #    'inventory_monitor_probed', False)
            inventory_dict['last_probed'] = inventory.custom_fields.get(
                'inventory_monitor_last_probe', "")
            inventory_dict['asset_numbers'] = inventory.custom_fields.get(
                'asset_numbers', "").split(';')
            inventory_dict['inventarizationNumber'] = inventory_dict['asset_numbers'][0]
            inventory_dict['active'] = inventory.custom_fields.get(
                'inventory_monitor_active', False)
            inventory_dict['device_name'] = device.name
            inventory_dict['name'] = inventory.name
            inventory_dict['site_name'] = site.name if site else ""
            inventory_dict['site_physical_address'] = site.physical_address if site else ""
            inventory_dict['site_longitude'] = site.longitude if site else ""
            inventory_dict['site_latitude'] = site.latitude if site else ""
            inventory_dict['site_location'] = location.name if location else ""
            if site and site['custom_fields'].get('komora_id', ""):
                inventory_dict['site_id'] = site['custom_fields'].get(
                    'komora_id', "")
            if device and device['custom_fields'].get('komora_id', ""):
                inventory_dict['device_id'] = device['custom_fields'].get(
                    'komora_id', "")
            if location and location.custom_fields.get('komora_id', ""):
                inventory_dict['location_id'] = location.custom_fields.get(
                    'komora_id', "")

            inventory_dict['netbox_id'] = inventory.id
            return inventory_dict

        # "asset_number": [items]
        result = {}

        for inv in NetboxConnection.get_connection().dcim.inventory_items.filter(cf_abra_discovered=True):
            inv_dict = inventory_to_json(inv)
            asset_numbers = []

            # if multiple asset_numbers found
            if ';' in inv.custom_fields.get('asset_numbers'):
                for asset_number in inv.custom_fields.get('asset_numbers').split(';'):
                    asset_numbers.append(asset_number)
            elif inv.custom_fields.get('asset_numbers') is not None:
                asset_numbers.append(inv.custom_fields.get('asset_numbers'))
            else:
                print("something went wrong", inv, inv.id)

            for asset_nmr in asset_numbers:
                # add inventory_item or init it
                if result.get(asset_nmr):
                    result[asset_nmr].append(inv_dict)
                else:
                    result[asset_nmr] = [inv_dict]

        return result



import logging
import os

import appdirs
import yaml

from komora_syncer import __appname__

logger = logging.getLogger(__name__)

os.environ["XDG_CONFIG_DIRS"] = "/etc"
CONFIG_DIRS = (
    appdirs.user_config_dir(__appname__),
    appdirs.site_config_dir(__appname__),
)
CONFIG_FILENAME = "config.yml"

def get_config():
    """
    Get config file and load it with yaml
    :returns: loaded config in yaml, as a dict object
    """
    
    if getattr(get_config, "cache", None):
        return get_config.cache

    if os.environ.get("CONFIG_FOLDER_PATH"):
        config_path = os.path.join(os.environ.get("CONFIG_FOLDER_PATH"), CONFIG_FILENAME)
    else:
        for d in CONFIG_DIRS:
            config_path = os.path.join(d, CONFIG_FILENAME)
            if os.path.isfile(config_path):
                break
        else:
            logger.error(
                "No configuration file can be found. Please create a "
                "config.yml in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
            exit(0)

    try:
        with open(config_path, "r") as config_file:
            conf = yaml.safe_load(config_file)
            get_config.cache = conf
            return conf
    except FileNotFoundError as e:
        logger.error(e)
        exit(0)


def setup_logger():
    logging_config = get_config().get("logging", {})
    level = logging_config.get('LOG_LEVEL', 'DEBUG')
    disable_existing_loggers = logging_config.get('DISABLE_EXISTING_LOGGERS', False)
    formatting = logging_config.get('LOG_FORMAT', '%(asctime)s - %(levelname)s - %(name)s - %(filename)s:%(lineno)d - %(message)s')

    # create logger
    logger = logging.getLogger(__appname__)
    logger.setLevel(level)

    # create console handler and set level
    ch = logging.StreamHandler()
    ch.setLevel(level)

    # create formatter
    formatter = logging.Formatter(formatting)

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)

    if disable_existing_loggers:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)




import json
import logging
import time

import requests

from komora_syncer.config import get_config
from komora_syncer.connections.KomoraConnection import KomoraConnection
from komora_syncer.helpers.utils import build_tenant_name
from komora_syncer.models.komora.contact import Contact
from komora_syncer.models.komora.device import Device
from komora_syncer.models.komora.district import District
from komora_syncer.models.komora.inventory import Inventory
from komora_syncer.models.komora.municipality import Municipality
from komora_syncer.models.komora.organization import Organization
from komora_syncer.models.komora.region import Region
from komora_syncer.models.komora.site import Site

# Log Warnings and higher errors from imported modules
logging.getLogger(requests.__name__).setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)
logger = logging.getLogger(__name__)


class KomoraApi:
    cache = {
        'sites': [],
        'regions': [],
        'districts': [],
        'municipalities': [],
        'organizations': [],
        'contacts': [],
        'devices': [],
        'inventories': [],
    }

    def __new__(self):
        raise TypeError("This is a static class")

    @staticmethod
    def store_cache(object_type, objects):
        KomoraApi.cache[object_type] = objects

    @staticmethod
    def get_object_instance(object_type, komora_data):
        object_map = {
            'sites': Site,
            'regions': Region,
            'districts': District,
            'municipalities': Municipality,
            'organizations': Organization,
            'devices': Device,
            'contacts': Contact,
            'inventories': Inventory,
        }
        if object_type not in object_map:
            raise ValueError(f"Unknown object type: {object_type}")
        return object_map[object_type](**komora_data)

    @staticmethod
    def check_cache(object_type):
        return KomoraApi.cache.get(object_type, [])

    @staticmethod
    def get_objects(object_type, api, filter_params=''):
        data = []

        try:
            logger.info(f"Polling {object_type} from Komora")
            for obj in KomoraConnection.get_records(api, filter_params=filter_params):
                data.append(KomoraApi.get_object_instance(object_type, obj))
        except Exception:
            logger.exception(f"Unable to poll {object_type} from Komora")
            raise

        return data

    @staticmethod
    def get_sites():
        sites = KomoraApi.check_cache('sites')
        if not sites:
            sites = KomoraApi.get_objects('sites', 'Site', '')
            KomoraApi.store_cache('sites', sites)
        return sites

    @staticmethod
    def get_organizations():
        organizations = KomoraApi.check_cache('organizations')
        if not organizations:
            organizations = KomoraApi.get_objects(
                'organizations', 'Organization', '')
            # Filter out duplicated CID and Name
            unique_cids_names = set()
            filtered_organizations = []
            for organization in organizations:
                cid_name = build_tenant_name(
                    organization.clientId, organization.name)
                if cid_name not in unique_cids_names:
                    unique_cids_names.add(cid_name)
                    filtered_organizations.append(organization)
                else:
                    logger.critical(f"Duplicated CID and Name: {cid_name}")
            KomoraApi.store_cache('organizations', filtered_organizations)
            organizations = filtered_organizations
        return organizations

    @staticmethod
    def get_regions():
        regions = KomoraApi.check_cache('regions')
        if not regions:
            regions = KomoraApi.get_objects('regions', 'Ruian/region')
            KomoraApi.store_cache('regions', regions)
        return regions

    @staticmethod
    def get_districts():
        districts = KomoraApi.check_cache('districts')
        if not districts:
            districts = KomoraApi.get_objects('districts', 'Ruian/district')
            KomoraApi.store_cache('districts', districts)
        return districts

    @staticmethod
    def get_municipalities():
        municipalities = KomoraApi.check_cache('municipalities')
        if not municipalities:
            municipalities = KomoraApi.get_objects(
                'municipalities', 'Ruian/municipality')
            KomoraApi.store_cache('municipalities', municipalities)
        return municipalities

    @staticmethod
    def get_devices():
        devices = KomoraApi.check_cache('devices')
        if not devices:
            filter_params = 'Filters.IsActive=true'
            devices = KomoraApi.get_objects('devices', 'Device', filter_params)
            KomoraApi.store_cache('devices', devices)
        return devices

    @staticmethod
    def post_devices(nb_devices):
        # Save nb_devices to file
        time_str = time.strftime("%Y%m%d-%H%M%S")
        filename = f"{time_str}_devices.json"
        filepath = get_config()['common']['DEVICES_DATA_PATH']
        file = filepath + filename

        with open(file, 'w') as outfile:
            json.dump(nb_devices, outfile)

        url = KomoraConnection.checkout_url("ServiceRecord/SendDeviceData")

        try:
            logger.info("Posting devices to Komora")
            requests.post(url, files={'dataFile': open(file, 'rb')})
        except requests.exceptions.RequestException:
            logger.exception("Unable to post devices to Komora")
            raise

    @staticmethod
    def get_contacts():
        contacts = KomoraApi.check_cache('contacts')
        if not contacts:
            contacts = KomoraApi.get_objects('contacts', 'Contact')
            KomoraApi.store_cache('contacts', contacts)
        return contacts

    @staticmethod
    def get_inventories():
        inventories = KomoraApi.check_cache('inventories')
        if inventories:
            return inventories

        inventories = KomoraApi.get_objects('inventories', 'Inventarization')
        KomoraApi.store_cache('inventories', inventories)
        return inventories

    @staticmethod
    def post_inventories(nb_inventories):
        api = "ServiceRecord/sendInventarizationData"
        url = KomoraConnection.checkout_url(api)

        time_str = time.strftime("%Y%m%d-%H%M%S")
        filename = f"{time_str}_inventarization.json"
        filepath = get_config()['common']['INVENTARIZATION_DATA_PATH']
        file = filepath + filename

        with open(file, 'w') as outfile:
            json.dump(nb_inventories, outfile)

        try:
            logger.info("Posting inventories to Komora")
            with open(file, 'rb') as data_file:
                requests.post(url, files={'dataFile': data_file})
        except Exception as e:
            logger.exception("Unable to post inventories to Komora")
            raise e


import logging
import traceback

import click

from komora_syncer.config import setup_logger
from komora_syncer.processors.Synchronizer import Synchronizer

SYNC_OPTIONS = ['all', 'organizations', 'regions', 'sites', 'devices', 'inventory']

# Initialize logging configuration
setup_logger()
logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@click.command()
@click.option('--sync', '-s', type=click.Choice(SYNC_OPTIONS), multiple=True, default=['all'], help="what should be synced")
@click.option('--site', '-t', type=click.STRING, help="Name of site to deploy")
def synchronize(sync, site):
    """
    Synchronize data between Netbox and Komora
    """
    synchronizer = Synchronizer()

    if site and 'sites' not in sync:
        logger.critical(
            "Site option '--site' is only available with sites 'synchronize --sync sites' synchronization")
        logger.critical("Exiting")
        return

    if site and 'sites' in sync:
        try:
            # Syncs Sites / Site, Location
            synchronizer.sync_sites(site_name=site)
            return
        except Exception as e:
            logger.error(f"Unable to synchronize site '{site}' - {e}")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'organizations' in sync:
        try:
            # Syncs Organizations / Tenants
            synchronizer.sync_organizations()
        except Exception as e:
            logger.error(f"Unable to synchronize organizations")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'regions' in sync:
        try:
            # Syncs Regions, Disctricts and Municipalities / Regions
            synchronizer.sync_regions()
        except Exception as e:
            logger.error(f"Unable to synchronize regions")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'sites' in sync:
        try:
            # Syncs Sites / Site, Location
            synchronizer.sync_sites()
        except Exception as e:
            logger.error(f"Unable to synchronize sites")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'devices' in sync:
        try:
            synchronizer.sync_devices()
        except Exception as e:
            logger.error(f"Unable to synchronize devices")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'inventory' in sync:
        try:
            synchronizer.sync_inventory_items()
        except Exception as e:
            logger.error(f"Unable to synchronize inventories")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return


cli.add_command(synchronize)


if __name__ == "__main__":
    # Display CLI
    cli()



class Inventory():
    def __init__(self, **kwargs):
        '''
           'active': True,
           'assetNumbers': 'T5',
           'deviceId': 9776,
           'deviceName': 'AS1-PRG',
           'deviceSerialNumber': 'FGL164510GR',
           'id': 55,
           'inventarizationNumber': 'T5',
           'lastProbed': '2017-01-01T00:00:00',
           'locationId': None,
           'locationName': None,
           'netboxUrl': 'www.seznam.cz',
           'siteId': 11,
           'siteName': 'Praha-CESNET'
        '''
        self.active = kwargs.get('active')
        self.assetNumbers = kwargs.get('assetNumbers')
        self.deviceId = kwargs.get('deviceId')
        self.deviceName = kwargs.get('deviceName')
        self.deviceSerialNumber = kwargs.get('deviceSerialNumber')
        self.id = kwargs.get('id')
        self.inventarizationNumber = kwargs.get('inventarizationNumber')
        self.lastProbed = kwargs.get('lastProbed')
        self.locationId = kwargs.get('locationId')
        self.locationName = kwargs.get('locationName')
        self.netboxUrl = kwargs.get('netboxUrl')
        self.siteId = kwargs.get('siteId')
        self.siteName = kwargs.get('siteName')

    def __str__(self):
        return self.inventarizationNumber

    def __repr__(self):
        return self.inventarizationNumber
